@@ -3,8 +3,8 @@ from __future__ import unicode_literals
 import logging
 import sys
 import mock
-import six
 
+from raven.utils.compat import iteritems, PY2
 from raven.base import Client
 from raven.handlers.logging import SentryHandler
 from raven.utils.stacks import iter_stack_frames
@@ -31,7 +31,7 @@ class LoggingIntegrationTest(TestCase):
     def make_record(self, msg, args=(), level=logging.INFO, extra=None, exc_info=None, name='root', pathname=__file__):
         record = logging.LogRecord(name, level, pathname, 27, msg, args, exc_info, 'make_record')
         if extra:
-            for key, value in six.iteritems(extra):
+            for key, value in iteritems(extra):
                 record.__dict__[key] = value
         return record
 
@@ -113,12 +113,20 @@ class LoggingIntegrationTest(TestCase):
 
         self.assertEqual(len(self.client.events), 1)
         event = self.client.events.pop(0)
-        if six.PY3:
+        if not PY2:
             expected = "'http://example.com'"
         else:
             expected = "u'http://example.com'"
         self.assertEqual(event['extra']['url'], expected)
 
+    def test_extra_data_dict_is_not_mutated(self):
+        # The code used to modify the dictionary included in extra arguments under the 'data' key.
+        # This is unexpected behavior, let's make sure it doesn't happen anymore.
+        data = {'data_key': 'data_value'}
+        record = self.make_record('irrelevant', extra={'data': data})
+        self.handler.emit(record)
+        self.assertEqual(data, {'data_key': 'data_value'})
+
     def test_logger_exc_info(self):
         try:
             raise ValueError('This is a test ValueError')
@@ -134,7 +142,7 @@ class LoggingIntegrationTest(TestCase):
 
         self.assertEqual(event['message'], 'This is a test info with an exception')
         assert 'exception' in event
-        exc = event['exception']['values'][0]
+        exc = event['exception']['values'][-1]
         self.assertEqual(exc['type'], 'ValueError')
         self.assertEqual(exc['value'], 'This is a test ValueError')
         self.assertTrue('sentry.interfaces.Message' in event)
@@ -151,7 +159,7 @@ class LoggingIntegrationTest(TestCase):
         self.assertEqual(event['message'], 'This is a test of args')
         msg = event['sentry.interfaces.Message']
         self.assertEqual(msg['message'], 'This is a test of %s')
-        expected = ("'args'",) if six.PY3 else ("u'args'",)
+        expected = ("'args'",) if not PY2 else ("u'args'",)
         self.assertEqual(msg['params'], expected)
 
     def test_record_stack(self):
@@ -207,7 +215,7 @@ class LoggingIntegrationTest(TestCase):
 
         self.assertEqual(len(self.client.events), 1)
         event = self.client.events.pop(0)
-        expected = "'foo'" if six.PY3 else "u'foo'"
+        expected = "'foo'" if not PY2 else "u'foo'"
         self.assertEqual(event['extra']['data'], expected)
 
     def test_tags(self):
@@ -229,6 +237,15 @@ class LoggingIntegrationTest(TestCase):
         event = self.client.events.pop(0)
         assert event['tags'] == {'foo': 'bar'}
 
+    def test_tags_merge(self):
+        handler = SentryHandler(self.client, tags={'foo': 'bar', 'biz': 'baz'})
+        record = self.make_record('Message', extra={'tags': {'foo': 'faz'}})
+        handler.emit(record)
+
+        self.assertEqual(len(self.client.events), 1)
+        event = self.client.events.pop(0)
+        assert event['tags'] == {'foo': 'faz', 'biz': 'baz'}
+
     def test_fingerprint_on_event(self):
         record = self.make_record('Message', extra={'fingerprint': ['foo']})
         self.handler.emit(record)
@@ -253,6 +270,17 @@ class LoggingIntegrationTest(TestCase):
         event = self.client.events.pop(0)
         assert event['server_name'] == 'foo'
 
+    def test_sample_rate(self):
+        record = self.make_record('Message', extra={'sample_rate': 0.0})
+        self.handler.emit(record)
+
+        self.assertEqual(len(self.client.events), 0)
+
+        record = self.make_record('Message', extra={'sample_rate': 1.0})
+        self.handler.emit(record)
+
+        self.assertEqual(len(self.client.events), 1)
+
 
 class LoggingHandlerTest(TestCase):
     def test_client_arg(self):
